import argparse
import functools
import pathlib

import numpy as np
import qpformat

from .._version import version

from . import definitions
from . import config

#: DryMass analysis output suffix (appended to data path)
OUTPUT_SUFFIX = "_dm"

META_MAPPER = {"medium index": ("medium index", 1),
               "pixel size um": ("pixel size", 1e6),
               "wavelength nm": ("wavelength", 1e9),
               }


@functools.lru_cache(maxsize=32)
def input_setting(path, section, key):
    """Ask the user for a configuration key"""
    cfg = config.ConfigFile(path)
    sec = cfg[section]
    description = definitions.config[section][key][2]
    if key not in sec or np.isnan(sec[key]):
        val = input("Please enter '{}' ({}): ".format(key, description))
        cfg.set_value("meta", key, val)


def main(path=None, req_meta=[]):
    """Main user dialog with optional "meta" kwargs required"""
    # get directories
    if path is None:
        path_in = parse().resolve()
    else:
        path_in = pathlib.Path(path).resolve()
    path_out = path_in.with_name(path_in.name + OUTPUT_SUFFIX)
    path_out.mkdir(exist_ok=True)
    # print directories (wrapped with functools so only executed once)
    print_info(path_in, path_out)
    # get known meta data kwargs from dataset
    transfer_meta_data(path_in, path_out)
    # user input missing meta data keyword values
    for mm in sorted(req_meta):
        input_setting(path=path_out,
                      section="meta",
                      key=mm)
    return path_in, path_out


@functools.lru_cache(maxsize=32)  # cached to avoid multiple prints
def parse():
    """Obtain the input data set path by parsing the command line"""
    print("DryMass version {}".format(version))
    parser = argparse.ArgumentParser(description='DryMass QPI analysis.')
    parser.add_argument('path', metavar='path', nargs='+', type=str,
                        help='Data path')
    args = parser.parse_args()
    # Workaround: We use nargs='+' and join the input to support white
    # spaces in path names.
    jpath = " ".join(args.path)
    path_in = pathlib.Path(jpath).resolve()
    return path_in


def transfer_meta_data(path_in, path_out):
    """Read input meta data and write it to the configuration file"""
    ds = qpformat.load_data(path=path_in)
    cfg = config.ConfigFile(path_out)
    sec = cfg["meta"]
    for key in sorted(META_MAPPER):
        dskey, mult = META_MAPPER[key]
        if (key not in sec or np.isnan(sec[key])) and dskey in ds.meta_data:
            cfg.set_value("meta", key, ds.meta_data[dskey] * mult)


@functools.lru_cache(maxsize=32)  # cached to avoid multiple prints
def print_info(path_in, path_out):
    """Print input and output paths"""
    print("Input:  {}".format(path_in))
    print("Output: {}".format(path_out))



import os
import tempfile
import shutil

import numpy as np
import qpimage

import drymass


def setup_test_data(radius=30, pxsize=1e-6, size=200, num=1, identifier=None,
                    medium_index=1.335, wavelength=550e-9, bg=None):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    image = (r < radius) * 1.3
    if bg is not None:
        image += bg
    qpi = qpimage.QPImage(data=image,
                          which_data="phase",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path = tempfile.mktemp(suffix=".h5", prefix="drymass_test_roi")
    dout = tempfile.mkdtemp(prefix="drymass_test_roi_")
    with qpimage.QPSeries(h5file=path, identifier=identifier) as qps:
        for ii in range(num):
            qpid = "{}_test_{}".format(identifier, ii)
            qps.add_qpimage(qpi, identifier=qpid)
    return qpi, path, dout


def test_basic():
    radius = 30
    pxsize = 1e-6
    qpi, path, dout = setup_test_data(radius=radius, pxsize=pxsize)
    path_out = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize)

    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert len(qpso) == 1
        qpi2 = qpso.get_qpimage(0)
        assert qpi != qpi2
        assert qpi.shape != qpi2.shape

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_corr_thresh():
    radius = 30
    pxsize = 1e-6
    bg = .3
    qpi, path, dout = setup_test_data(radius=radius, pxsize=pxsize, bg=bg)

    bg_pha_kw = {"fit_offset": "mean",
                 "fit_profile": "tilt",
                 "border_perc": 0,
                 "border_px": 0}
    path_out = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   bg_pha_kw=bg_pha_kw,
                                   bg_pha_bin=bg*1.1)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert np.min(qpso[0].pha) == 0
        assert np.allclose(np.max(qpso[0].pha), np.max(qpi.pha) - bg)

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_corr_mask():
    radius = 30
    pxsize = 1e-6
    size = 200
    bg = np.zeros((size, size)) + .1
    bg += np.linspace(-.3, .5, size).reshape(-1, 1)
    _, path, dout = setup_test_data(radius=radius, size=size,
                                    pxsize=pxsize, bg=bg)
    qpiref, p2, d2 = setup_test_data(radius=radius, size=size,
                                     pxsize=pxsize, bg=None)

    bg_pha_kw = {"fit_offset": "mean",
                 "fit_profile": "tilt",
                 "border_perc": 0,
                 "border_px": 0}
    path_out = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   bg_pha_kw=bg_pha_kw,
                                   bg_pha_mask_radial_clearance=1.1)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        assert np.allclose(np.min(qpso[0].pha), 0, atol=4e-9, rtol=0)
        assert np.allclose(np.max(qpso[0].pha), np.max(qpiref.pha),
                           atol=1.1e-7, rtol=0)

    try:
        os.remove(path)
        os.remove(p2)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)
    shutil.rmtree(d2, ignore_errors=True)


def test_ret_changed():
    radius = 30
    pxsize = 1e-6
    _qpi, path, dout = setup_test_data(radius=radius, pxsize=pxsize)
    _p1, ch1 = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   ret_changed=True)
    _p2, ch2 = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   ret_changed=True)
    assert ch1, "First call should create data on disk"
    assert not ch2, "Second call should reuse data on disk"

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_no_search():
    radius = 30
    pxsize = 1e-6
    identifier = "asdkn179"
    _qpi, path, dout = setup_test_data(radius=radius,
                                       pxsize=pxsize,
                                       identifier=identifier)
    _p1, rm1 = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   ret_roimgr=True)
    _p2, rm2 = drymass.extract_roi(path,
                                   dir_out=dout,
                                   size_m=2*radius*pxsize,
                                   search_enabled=False,
                                   ret_roimgr=True,)

    assert rm1.rois == rm2.rois
    assert rm1.identifier == identifier
    assert rm2.identifier == identifier

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from drymass import cli
import imageio
import matplotlib.pylab as plt
import numpy as np

from t02_method_comparison import dot_boxplot

path = "DHM_HL60_cells.zip"

# setup output directory
path_in, path_out = cli.dialog.main(path)

# set metadata
cfg = cli.config.ConfigFile(path_out)
cfg.set_value("meta", "medium index", 1.335)
cfg.set_value("meta", "pixel size um", 0.107)
cfg.set_value("meta", "wavelength nm", 633)

# extract ROIs
cli.cli_extract_roi(path=path_in)

# get image 27
# Note: Remove the DHM_HL60_cells.zip_dm folder to reproduce the figure
senpath = path_out / cli.FILE_SENSOR_WITH_ROI_IMAGE
sentif = imageio.mimread(senpath)[26]
imageio.imsave("_t02_roi_search1.jpg", sentif[..., :3])

# rerun with updated configuration
cfg.set_value("specimen", "size um", 13)
cfg.set_value("roi", "pad border px", 80)
cfg.set_value("roi", "size variation", .2)
cfg.set_value("roi", "exclude overlap px", 100)
cli.cli_extract_roi(path=path_in)

# remove ROI slices
cfg.set_value("roi", "ignore data", ["8.4", "15.2", "18.2", "18.3", "35.2"])

# bg correction
cfg.set_value("bg", "phase border px", 30)
cfg.set_value("bg", "phase profile", "poly2o")

# edge projection
cfg.set_value("sphere", "method", "edge")
cfg.set_value("sphere", "model", "projection")
cli.cli_analyze_sphere(path=path_in)

senpath = path_out / cli.FILE_SPHERE_ANALYSIS_IMAGE.format("edge",
                                                           "projection")
sentif = imageio.mimread(senpath)[52]
imageio.imsave("_t02_edge_projection.jpg", sentif[..., :3])

# image projection
cfg.set_value("sphere", "method", "image")
cfg.set_value("sphere", "model", "projection")
cli.cli_analyze_sphere(path=path_in)

# image rytov
cfg.set_value("sphere", "method", "image")
cfg.set_value("sphere", "model", "rytov")
cli.cli_analyze_sphere(path=path_in)

# image rytov-sc
cfg.set_value("sphere", "method", "image")
cfg.set_value("sphere", "model", "rytov-sc")
cli.cli_analyze_sphere(path=path_in)

senpath = path_out / cli.FILE_SPHERE_ANALYSIS_IMAGE.format("image",
                                                           "rytov-sc")
sentif = imageio.mimread(senpath)[52]
imageio.imsave("_t02_image_rytov-sc.jpg", sentif[..., :3])

# comparison plot
ri_data = [
    np.loadtxt(path_out / "sphere_image_rytov-sc_statistics.txt",
               usecols=(1,)),
    np.loadtxt(path_out / "sphere_image_rytov_statistics.txt",
               usecols=(1,)),
    np.loadtxt(path_out / "sphere_image_projection_statistics.txt",
               usecols=(1,)),
    np.loadtxt(path_out / "sphere_edge_projection_statistics.txt",
               usecols=(1,)),
    ]
colors = ["#E48620", "#DE2400", "#6e559d", "#048E00"]
labels = ["image rytov-sc", "image rytov",
          "image projection", "edge projection"]

plt.figure(figsize=(8, 5))
ax = plt.subplot(111, title="HL60 (DHM)")
ax.set_ylabel("refractive index")
dot_boxplot(ax=ax, data=ri_data, colors=colors, labels=labels)
plt.tight_layout()
plt.savefig("_t02_reproduced_5d.jpg")
plt.close()



"""Utility methods"""
import hashlib
import pathlib

import numpy as np


def hash_file(path, blocksize=65536):
    """Compute sha256 hex-hash of a file

    Parameters
    ----------
    path: str
        path to the file
    blocksize: int
        block size read from the file

    Returns
    -------
    hex: str
        The first six characters of the hash
    """
    fname = pathlib.Path(path)
    hasher = hashlib.sha256()
    with fname.open('rb') as fd:
        buf = fd.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
    return hasher.hexdigest()[:6]


def hash_object(obj):
    """Compute sha256 hex-hash of a Python object

    Objects in dicts/lists/tuples are joined together before hashing
    using :func:`obj2bytes` in this module.

    Returns
    -------
    hex: str
        The first six characters of the hash
    """
    ihasher = hashlib.sha256()
    ihasher.update(obj2bytes(obj))
    return ihasher.hexdigest()[:6]


def obj2bytes(obj):
    """String representation of an object for hashing"""
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, (bool, int, float)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tostring()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(list(obj.items())))
    elif hasattr(obj, "identifier"):
        return obj2bytes(obj.identifier)
    else:
        raise ValueError("No rule to convert object '{}' to bytes.".
                         format(obj.__class__))



import matplotlib.pylab as plt
import numpy as np


def dot_boxplot(ax, data, colors, labels, **kwargs):
    """Combined box and scatter plot"""
    box_list = []

    for ii in range(len(data)):
        # set same random state for every scatter plot
        rs = np.random.RandomState(42).get_state()
        np.random.set_state(rs)
        y = data[ii]
        x = np.random.normal(ii+1, 0.15, len(y))
        plt.plot(x, y, 'o', alpha=0.5, color=colors[ii])
        box_list.append(y)

    ax.boxplot(box_list,
               sym="",
               medianprops={"color": "black", "linestyle": "solid"},
               widths=0.3,
               labels=labels,
               **kwargs)
    plt.grid(axis="y")


if __name__ == "__main__":
    ri_data = [
        np.loadtxt("sphere_image_rytov-sc_statistics.txt", usecols=(1,)),
        np.loadtxt("sphere_image_rytov_statistics.txt", usecols=(1,)),
        np.loadtxt("sphere_image_projection_statistics.txt", usecols=(1,)),
        np.loadtxt("sphere_edge_projection_statistics.txt", usecols=(1,)),
        ]
    colors = ["#E48620", "#DE2400", "#6e559d", "#048E00"]
    labels = ["image rytov-sc", "image rytov",
              "image projection", "edge projection"]

    plt.figure(figsize=(8, 5))
    ax = plt.subplot(111, title="HL60 (DHM)")
    ax.set_ylabel("refractive index")
    dot_boxplot(ax=ax, data=ri_data, colors=colors, labels=labels)
    plt.tight_layout()
    plt.show()



import pathlib
import shutil
import tempfile

import numpy as np
import pytest
import qpimage

from drymass.cli import cli_extract_roi, config, dialog
from drymass.extractroi import FILE_ROI_DATA_TIF, FILE_SLICES


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path_in = tempfile.mktemp(suffix=".h5", prefix="drymass_test_cli_convert")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tt") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="image_{}".format(ii))
    # add drymass configuration file
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
    cfg.set_value(section="meta", key="wavelength nm", value=wavelength*1e9)
    cfg.set_value(section="meta", key="medium index", value=medium_index)
    cfg.set_value(section="specimen", key="size um",
                  value=radius_px*2*pxsize*1e6)
    return qpi, path_in, path_out


def test_base():
    _, path_in, path_out = setup_test_data(num=2)
    h5data = cli_extract_roi(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert len(qps) == 2
        # This might fail when the algorithms or default parameters for
        # finding ROIs change:
        assert qps[0].shape == (139, 139)

    # check existence of files
    pathtif = path_out / FILE_ROI_DATA_TIF
    assert pathtif.exists()
    assert pathtif.stat().st_size > 1500

    pathsl = path_out / FILE_SLICES
    assert pathsl.exists()
    assert len(pathsl.read_bytes()) > 100

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_exclude_roi():
    _, path_in, path_out = setup_test_data(num=2)
    cli_extract_roi(path=path_in)
    cfg = config.ConfigFile(path_out)
    h5data = cli_extract_roi(path=path_in, ret_data=True)
    with qpimage.QPSeries(h5file=h5data) as qps:
        assert len(qps) == 2

    # remove first image
    cfg.set_value(section="roi", key="ignore data", value="1.1")
    h5data = cli_extract_roi(path=path_in, ret_data=True)
    with qpimage.QPSeries(h5file=h5data) as qps:
        assert len(qps) == 1

    # remove second image
    cfg.set_value(section="roi", key="ignore data", value="2")
    h5data = cli_extract_roi(path=path_in, ret_data=True)
    with qpimage.QPSeries(h5file=h5data) as qps:
        assert len(qps) == 1

    # remove all
    cfg.set_value(section="roi", key="ignore data", value=["1", "2"])
    h5data = cli_extract_roi(path=path_in, ret_data=True)
    with qpimage.QPSeries(h5file=h5data) as qps:
        assert len(qps) == 0

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_exclude_roi_bad():
    _, path_in, path_out = setup_test_data(num=2)
    cli_extract_roi(path=path_in)
    cfg = config.ConfigFile(path_out)
    h5data = cli_extract_roi(path=path_in, ret_data=True)
    with qpimage.QPSeries(h5file=h5data) as qps:
        assert len(qps) == 2

    # bad image
    cfg.set_value(section="roi", key="ignore data", value="3")
    try:
        cli_extract_roi(path=path_in, ret_data=True)
    except ValueError:
        pass
    else:
        assert False

    # bad roi
    cfg.set_value(section="roi", key="ignore data", value="1.2")
    try:
        cli_extract_roi(path=path_in, ret_data=True)
    except ValueError:
        pass
    else:
        assert False

    # bad image and roi
    cfg.set_value(section="roi", key="ignore data", value="4.2")
    try:
        cli_extract_roi(path=path_in, ret_data=True)
    except ValueError:
        pass
    else:
        assert False

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_reuse():
    _, path_in, path_out = setup_test_data(num=2)
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="meta", key="pixel size um", value=1)

    h5data = cli_extract_roi(path=path_in, ret_data=True)

    time = h5data.stat().st_mtime

    # Do the same thing
    cli_extract_roi(path=path_in, ret_data=True)
    assert time == h5data.stat().st_mtime

    # Change something
    cfg.set_value(section="meta", key="pixel size um", value=1.01)
    cli_extract_roi(path=path_in, ret_data=True)
    assert time != h5data.stat().st_mtime

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_no_roi_found():
    _, path_in, _ = setup_test_data(radius_px=0)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        # catches sys.exit() calls
        cli_extract_roi(path=path_in, ret_data=True)
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 1


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from drymass import cli
import imageio
import numpy as np


path = "QLSR_PAA_beads"

# setup output directory
path_in, path_out = cli.dialog.main(path)

# set metadata
cfg = cli.config.ConfigFile(path_out)
cfg.set_value("meta", "medium index", 1.335)
cfg.set_value("meta", "pixel size um", 0.14)
cfg.set_value("meta", "wavelength nm", 647)

# perform analysis
cli.cli_analyze_sphere(path=path_in)

# get first sensor image
senpath = path_out / cli.FILE_SENSOR_WITH_ROI_IMAGE
sentif = imageio.imread(senpath)
imageio.imsave("_t01_sensor_roi_image.jpg", sentif[..., :3])

# get first sphere image
roipath = path_out / cli.FILE_SPHERE_ANALYSIS_IMAGE.format("edge",
                                                           "projection")
roitif = imageio.imread(roipath)
imageio.imsave("_t01_sphere_edge_projection_image.jpg", roitif[..., :3])


cfg.set_value("roi", "ignore data", ["4.1", "7.1", "24.1", "26.1", "27.1",
                                     "35.1", "36.1", "39.1", "40.1", "51.1",
                                     "52.1", "55.1", "58.1", "60.1", "64.1",
                                     "67.1", "71.1"])
# perform analysis again
cli.cli_analyze_sphere(path=path_in)

# extract mean and SD of RI
ri = np.loadtxt(path_out / "sphere_edge_projection_statistics.txt",
                usecols=(1,))
print("average: ", np.average(ri))
print("standard deviation: ", np.std(ri))



from os import fspath
import pathlib
import warnings

import numpy as np
import qpimage
import qpsphere
from skimage.external import tifffile
import skimage.filters

from . import search, util
from .roi import ROIManager


#: Default background correction keyword arguments
BG_DEFAULT_KW = {"fit_offset": "mean",
                 "fit_profile": "tilt",
                 "border_perc": 5,
                 "border_px": 5,
                 }
#: Output ROI qpimage.QPSeries data
FILE_ROI_DATA_H5 = "roi_data.h5"
#: Output phase/amplitude TIFF data
FILE_ROI_DATA_TIF = "roi_data.tif"
#: Output slice locations
FILE_SLICES = "roi_slices.txt"


def _bg_correct(qpi, which_data, bg_kw={}, bg_mask_thresh=np.nan,
                bg_mask_sphere_kw={}):
    if bg_kw:
        if isinstance(bg_mask_thresh, str) or not np.isnan(bg_mask_thresh):
            if which_data == "phase":
                image = qpi.pha
            else:
                image = qpi.amp
            mask1 = image2mask(image,
                               value_or_method=bg_mask_thresh)
        else:
            mask1 = None
        if not np.isnan(bg_mask_sphere_kw["radial_clearance"]):  # sphere mask
            mask2 = qpsphere.cnvnc.bg_phase_mask_for_qpi(
                qpi=qpi,
                r0=bg_mask_sphere_kw["r0"],
                method="edge",
                model="projection",
                edgekw=bg_mask_sphere_kw["edgekw"],
                imagekw={},
                radial_clearance=bg_mask_sphere_kw["radial_clearance"])
        else:
            mask2 = None
        # combine masks
        if mask1 is None and mask2 is None:
            mask = None
        elif mask1 is None:
            mask = mask2
        elif mask2 is None:
            mask = mask1
        else:
            mask = np.logical_and(mask1, mask2)
        # perforn actual bg correction
        qpi.compute_bg(which_data=which_data,
                       from_mask=mask,
                       **bg_kw)


def _extract_roi(h5in, h5out, slout, imout, size_m, size_var, max_ecc,
                 dist_border, pad_border, exclude_overlap, ignore_data,
                 bg_amp_kw, bg_amp_bin, bg_amp_mask_sphere_kw,
                 bg_pha_kw, bg_pha_bin, bg_pha_mask_sphere_kw,
                 search_enabled):
    # Determine ROI location
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
        rmgr = ROIManager(qps.identifier)
        if search_enabled:
            for ii in range(len(qps)):
                # new indexing convention in drymass 0.6.0
                image_index = ii + 1
                qpi = qps[ii]
                # find objects
                slices = search.search_phase_objects(
                    qpi=qpi,
                    size_m=size_m,
                    size_var=size_var,
                    max_ecc=max_ecc,
                    dist_border=dist_border,
                    pad_border=pad_border,
                    exclude_overlap=exclude_overlap)
                for jj, sl in enumerate(slices):
                    # new indexing convention in drymass 0.6.0
                    roi_index = jj + 1
                    slident = "{}.{}".format(qpi["identifier"], roi_index)
                    rmgr.add(roi_slice=sl,
                             image_index=image_index,
                             roi_index=roi_index,
                             identifier=slident)
            rmgr.save(slout)
        else:
            rmgr.load(slout)

    # Verify ignore_data parameter
    if ignore_data:
        bad_ignore = []
        for item in ignore_data:
            if item.count("."):
                roims = rmgr.get_from_image_index(int(item.split(".")[0]))
                roims = ["{}.{}".format(r.image_index,
                                        r.roi_index) for r in roims]
                if item not in roims:
                    bad_ignore.append(item)
            else:
                if not rmgr.get_from_image_index(int(item)):
                    bad_ignore.append(item)
        if bad_ignore:
            msg = "The following ROIs/images are not present but are set in " \
                  + "`ignore_data`: {}".format(", ".join(bad_ignore))
            raise ValueError(msg)

    # Extract ROI images
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps, \
            qpimage.QPSeries(h5file=h5out, h5mode="w") as qps_roi, \
            tifffile.TiffWriter(fspath(imout), imagej=True) as tf:
        for ii in range(len(qps)):
            # new indexing convention in drymass 0.6.0
            image_index = ii + 1
            # image to analyze
            qpi = qps[ii]
            # available ROIs
            rois = rmgr.get_from_image_index(image_index)
            for jj, roi in enumerate(rois):
                # new indexing convention in drymass 0.6.0
                roi_index = jj + 1
                if is_ignored_roi(roi=roi, ignore_data=ignore_data):
                    # ignore data
                    continue
                # Extract the ROI
                qpisl = qpi.__getitem__(roi.roi_slice)
                # amplitude bg correction
                _bg_correct(qpi=qpisl,
                            which_data="amplitude",
                            bg_kw=bg_amp_kw,
                            bg_mask_thresh=bg_amp_bin,
                            bg_mask_sphere_kw=bg_amp_mask_sphere_kw)
                # phase bg correction
                _bg_correct(qpi=qpisl,
                            which_data="phase",
                            bg_kw=bg_pha_kw,
                            bg_mask_thresh=bg_pha_bin,
                            bg_mask_sphere_kw=bg_pha_mask_sphere_kw)
                slident = "{}.{}".format(qpi["identifier"], roi_index)
                if roi.identifier != slident:
                    # This might happen if the user does not know the
                    # image identifier and builds his own `FILE_SLICES`.
                    msg = "Mismatch of slice and QPImage identifiers: " \
                          + "{} vs {}!".format(roi.identifier, slident)
                    warnings.warn(msg)
                    # override `slident` with user identifier
                    slident = roi.identifier
                qps_roi.add_qpimage(qpisl, identifier=slident)

        if len(qps_roi):
            # Write TIF
            # determine largest image
            sxmax = np.max([qq.shape[0] for qq in qps_roi])
            symax = np.max([qq.shape[1] for qq in qps_roi])
            dummy = np.zeros((2, sxmax, symax), dtype=np.float32)
            for qpir in qps_roi:
                dummy[0, :, :] = 0
                dummy[1, :, :] = 1
                res = 1 / qpir["pixel size"] * 1e-6  # use µm
                sx, sy = qpir.shape
                dummy[0, :sx, :sy] = qpir.pha
                dummy[1, :sx, :sy] = qpir.amp
                tf.save(data=dummy, resolution=(res, res, None), compress=9)
    return rmgr


def extract_roi(h5series, dir_out, size_m, size_var=.5, max_ecc=.7,
                dist_border=10, pad_border=40, exclude_overlap=30.,
                ignore_data=None,
                bg_amp_kw=BG_DEFAULT_KW, bg_amp_bin=np.nan,
                bg_amp_mask_radial_clearance=np.nan,
                bg_pha_kw=BG_DEFAULT_KW, bg_pha_bin=np.nan,
                bg_pha_mask_radial_clearance=np.nan,
                bg_sphere_edge_kw={}, search_enabled=True,
                ret_roimgr=False, ret_changed=False):
    """Extract ROIs from a qpimage.QPSeries hdf5 file

    Parameters
    ----------
    h5series: str
        Path of qpimage.QPSeries hdf5 file
    dir_out: str
        Path to output directory
    size_m: float
        Approximate diameter of the specimen [m]
    size_var: float
        Allowed variation relative to specimen size
    max_ecc: float
        Allowed maximal eccentricity of the specimen
    dist_border: int
        Minimum distance of objects to image border [px]
    pad_border: int
        Padding of object regions [px]
    exclude_overlap: float
        Allowed distance between two objects [px]
    ignore_data: list of str
        Identifiers for sensor images or ROIs to be excluded from
        further analysis. These will be labeled in the output
        tiff file and not written to the output qpseries file.
    bg_amp_kw: dict or None
        Amplitude image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_amp_bin: float or str
        The amplitude binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    bg_amp_mask_radial_clearance: float
        If not NaN, use :func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`
        to compute a mask image and use it for amplitude
        background correction.
    bg_pha_kw: dict or None
        Phase image background correction keyword arguments
        (see :func:`qpimage.QPImage.compute_bg`), defaults
        to `BG_DEFAULT_KW`, set to `None` to disable correction
    bg_pha_bin: float or str
        The phase binary threshold value or the method for binary
        threshold determination; see :mod:`skimage.filters`
        `threshold_*` methods
    bg_pha_mask_radial_clearance: float
        If not NaN, use :func:`qpsphere.cnvnc.bg_phase_mask_for_qpi`
        to compute a mask image and use it for phase
        background correction.
    search_enabled: bool
        If True, perform automated search for ROIs using the
        parameters above. If False, extract the ROIs from `FILE_SLICES`
        and only perform background correction using the `bg_*`
        parameters.
    ret_roimgr: bool
        Return the ROIManager instance of the found ROIs
    ret_changed: bool
        Return boolean indicating whether the ROI data on disk was
        created/updated (True) or whether previously created ROI
        data was used (False).

    Notes
    -----
    The output hdf5 file `dir_out/FILE_ROI_DATA_H5` is a
    :class:`qpimage.QPSeries` file with the keyword "identifier"
    consisting of two hashes: one from the relevant arguments
    to this method and one from the file `dir_out/FILE_SLICES`.
    This is to ensure that user-manipulated data is taken into
    account when deciding whether the ROIs should be re-computed
    after an initial run.
    """
    h5in = pathlib.Path(h5series)
    dout = pathlib.Path(dir_out)

    h5out = dout / FILE_ROI_DATA_H5
    imout = dout / FILE_ROI_DATA_TIF
    slout = dout / FILE_SLICES

    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
        cfgid = util.hash_object([qps,
                                  size_m,
                                  size_var,
                                  max_ecc,
                                  dist_border,
                                  pad_border,
                                  exclude_overlap,
                                  ignore_data,
                                  bg_amp_kw,
                                  bg_amp_bin,
                                  bg_amp_mask_radial_clearance,
                                  bg_pha_kw,
                                  bg_pha_bin,
                                  bg_pha_mask_radial_clearance,
                                  bg_sphere_edge_kw,
                                  search_enabled,
                                  ])

    # Determine whether we have to extract the ROIs
    if h5out.exists() and slout.exists():
        slid = util.hash_file(slout)
        identifier = "{}-{}".format(cfgid, slid)
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qpo:
            if qpo.identifier == identifier:
                create = False
            else:
                create = True
    else:
        create = True

    if create:
        bg_amp_mask_sphere_kw = {
            "r0": size_m / 2,
            "edgekw": bg_sphere_edge_kw,
            "radial_clearance": bg_amp_mask_radial_clearance}
        bg_pha_mask_sphere_kw = {
            "r0": size_m / 2,
            "edgekw": bg_sphere_edge_kw,
            "radial_clearance": bg_pha_mask_radial_clearance}

        rmgr = _extract_roi(
            h5in=h5in,
            h5out=h5out,
            slout=slout,
            imout=imout,
            size_m=size_m,
            size_var=size_var,
            max_ecc=max_ecc,
            dist_border=dist_border,
            pad_border=pad_border,
            exclude_overlap=exclude_overlap,
            ignore_data=ignore_data,
            bg_amp_kw=bg_amp_kw,
            bg_amp_bin=bg_amp_bin,
            bg_amp_mask_sphere_kw=bg_amp_mask_sphere_kw,
            bg_pha_kw=bg_pha_kw,
            bg_pha_bin=bg_pha_bin,
            bg_pha_mask_sphere_kw=bg_pha_mask_sphere_kw,
            search_enabled=search_enabled,
            )
        # manually set the identifier with the updated file slout
        slid = util.hash_file(slout)
        identifier = "{}-{}".format(cfgid, slid)
        with qpimage.QPSeries(h5file=h5out, h5mode="a") as qpo:
            qpo.h5.attrs["identifier"] = identifier
    else:
        with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps:
            rmgr = ROIManager(qps.identifier)
        rmgr.load(slout)

    ret = [h5out]
    if ret_roimgr:
        ret.append(rmgr)
    if ret_changed:
        ret.append(create)
    if len(ret) == 1:
        ret = ret[0]
    return ret


def image2mask(image, value_or_method):
    """Convert an image to a binary mask

    Parameters
    ----------
    image: 2d np.ndarray
        Input image
    value_or_method: float or str
        Either a threshold value or a string naming a
        filter method in :mod:`skimage.filters`.
    """
    if isinstance(value_or_method, str):
        method = getattr(skimage.filters, value_or_method)
        return image < method(image)
    else:
        return image < value_or_method


def is_ignored_roi(roi, ignore_data):
    """Determine whether a specific ROI should be ignored

    Parameters
    ----------
    roi: drymass.roi.ROI
        ROI instance
    ignore_data: list of str
        List of strings of the form "image_index" or
        "image_index.roi_index" that identify ROIs that
        should be ignored. For instance
        ["1.0", "2.1", "3"].
    """
    imid = roi.image_index
    roid = roi.roi_index
    if (ignore_data
        and (str(imid) in ignore_data
             or "{}.{}".format(imid, roid) in ignore_data)):
        return True
    else:
        return False



import io
from os import fspath
import pathlib
import sys

import matplotlib.image as mpimg
import numpy as np
import qpimage
from skimage.external import tifffile

from . import config
from . import dialog
from . import plot

from ..anasphere import analyze_sphere
from ..converter import convert
from ..extractroi import extract_roi

#: Matplotlib images of sensor data with labeled ROI (TIFF)
FILE_SENSOR_WITH_ROI_IMAGE = "sensor_roi_images.tif"
#: Matplotlib images of sphere analysis (TIFF)
FILE_SPHERE_ANALYSIS_IMAGE = "sphere_{}_{}_images.tif"


def cli_analyze_sphere(path=None, ret_data=False):
    """Perform sphere analysis"""
    path_in, path_out = dialog.main(path=path,
                                    req_meta=["medium index",
                                              "pixel size um",
                                              "wavelength nm"])
    cfg = config.ConfigFile(path_out)
    h5roi = cli_extract_roi(path=path_in, ret_data=True)
    print("Performing sphere analysis... ", end="", flush=True)
    # canny edge detection parameters
    edgekw = {
        "clip_rmin": cfg["sphere"]["edge clip radius min"],
        "clip_rmax": cfg["sphere"]["edge clip radius max"],
        "mult_coarse": cfg["sphere"]["edge coarse"],
        "mult_fine": cfg["sphere"]["edge fine"],
        "maxiter": cfg["sphere"]["edge iter"],
    }
    # image fitting parameters
    imagekw = {
        "crel": cfg["sphere"]["image fit range position"],
        "rrel": cfg["sphere"]["image fit range radius"],
        "nrel": cfg["sphere"]["image fit range refractive index"],
        "fix_pha_offset": cfg["sphere"]["image fix phase offset"],
        "max_iter": cfg["sphere"]["image iter"],
        "stop_dc": cfg["sphere"]["image stop delta position"],
        "stop_dr": cfg["sphere"]["image stop delta radius"],
        "stop_dn": cfg["sphere"]["image stop delta refractive index"],
        "verbose": cfg["sphere"]["image verbosity"],
    }
    h5sim, changed = analyze_sphere(
        h5roi=h5roi,
        dir_out=path_out,
        r0=cfg["specimen"]["size um"] / 2 * 1e-6,
        method=cfg["sphere"]["method"],
        model=cfg["sphere"]["model"],
        alpha=cfg["sphere"]["refraction increment"],
        rad_fact=cfg["sphere"]["radial inclusion factor"],
        edgekw=edgekw,
        imagekw=imagekw,
        ret_changed=True,
        )

    print("Done.")

    if changed and cfg["output"]["sphere images"]:
        print("Plotting sphere images... ", end="", flush=True)
        tifout = path_out / FILE_SPHERE_ANALYSIS_IMAGE.format(
            cfg["sphere"]["method"],
            cfg["sphere"]["model"]
        )
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps_roi, \
                qpimage.QPSeries(h5file=h5sim, h5mode="r") as qps_sim, \
                tifffile.TiffWriter(fspath(tifout), imagej=True) as tf:
            for ii in range(len(qps_roi)):
                qpi_real = qps_roi[ii]
                qpi_sim = qps_sim[ii]
                assert qpi_real["identifier"] in qpi_sim["identifier"]
                imio = io.BytesIO()
                plot.plot_qpi_sphere(qpi_real=qpi_real,
                                     qpi_sim=qpi_sim,
                                     path=imio,
                                     simtype=cfg["sphere"]["model"])
                imio.seek(0)
                imdat = (mpimg.imread(imio) * 255).astype("uint8")
                tf.save(imdat, compress=9)
        print("Done")

    if ret_data:
        return h5sim


def cli_convert(path=None, ret_data=False):
    """Convert input data to QPSeries data"""
    path_in, path_out = dialog.main(path=path,
                                    req_meta=["pixel size um",
                                              "wavelength nm"])
    cfg = config.ConfigFile(path_out)
    print("Converting input data... ", end="", flush=True)
    meta_data = {"pixel size": cfg["meta"]["pixel size um"] * 1e-6,
                 "wavelength": cfg["meta"]["wavelength nm"] * 1e-9,
                 "medium index": cfg["meta"]["medium index"],
                 }

    holo_kw = {"filter_name": cfg["holo"]["filter name"],
               "filter_size": cfg["holo"]["filter size"],
               "sideband": cfg["holo"]["sideband"],
               }

    bg_data_amp = parse_bg_value(cfg["bg"]["amplitude data"],
                                 reldir=path_in.parent)
    bg_data_pha = parse_bg_value(cfg["bg"]["phase data"],
                                 reldir=path_in.parent)

    h5series, ds, changed = convert(path_in=path_in,
                                    dir_out=path_out,
                                    meta_data=meta_data,
                                    holo_kw=holo_kw,
                                    bg_data_amp=bg_data_amp,
                                    bg_data_pha=bg_data_pha,
                                    write_tif=cfg["output"]["sensor tif data"],
                                    ret_dataset=True,
                                    ret_changed=True,
                                    )
    if "hologram" not in ds.storage_type:
        # remove "holo" section from configuration
        cfg.remove_section("holo")
    if changed:
        print("Done.")
    else:
        print("Reusing.")
    if ret_data:
        return h5series


def cli_extract_roi(path=None, ret_data=False):
    """Extract regions of interest"""
    path_in, path_out = dialog.main(path=path)
    # cli_convert will ask for the required meta data
    h5series = cli_convert(path=path_in, ret_data=True)
    # get the configuration after cli_convert was run
    cfg = config.ConfigFile(path_out)
    print("Extracting ROIs... ", end="", flush=True)
    # background correction
    if cfg["bg"]["enabled"]:
        bg_amp_kw = {"fit_offset": cfg["bg"]["amplitude offset"],
                     "fit_profile": cfg["bg"]["amplitude profile"],
                     "border_perc": cfg["bg"]["amplitude border perc"],
                     "border_px": cfg["bg"]["amplitude border px"],
                     }
        bg_pha_kw = {"fit_offset": cfg["bg"]["phase offset"],
                     "fit_profile": cfg["bg"]["phase profile"],
                     "border_perc": cfg["bg"]["phase border perc"],
                     "border_px": cfg["bg"]["phase border px"],
                     }
        # Only get the Canny edge detection parameters if needed
        if not (np.isnan(cfg["bg"]["phase mask sphere"])
                and np.isnan(cfg["bg"]["amplitude mask sphere"])):
            dialog.main(path=path, req_meta=["medium index"])
            # This will generate the [sphere] section in drymass.cfg
            edge_kw = {
                "clip_rmin": cfg["sphere"]["edge clip radius min"],
                "clip_rmax": cfg["sphere"]["edge clip radius max"],
                "mult_coarse": cfg["sphere"]["edge coarse"],
                "mult_fine": cfg["sphere"]["edge fine"],
                "maxiter": cfg["sphere"]["edge iter"],
            }
        else:
            edge_kw = {}
    else:
        bg_amp_kw = None
        bg_pha_kw = None
        edge_kw = {}

    h5roi, rmgr, changed = extract_roi(
        h5series=h5series,
        dir_out=path_out,
        size_m=cfg["specimen"]["size um"] * 1e-6,
        size_var=cfg["roi"]["size variation"],
        max_ecc=cfg["roi"]["eccentricity max"],
        dist_border=cfg["roi"]["dist border px"],
        pad_border=cfg["roi"]["pad border px"],
        exclude_overlap=cfg["roi"]["exclude overlap px"],
        ignore_data=cfg["roi"]["ignore data"],
        bg_amp_kw=bg_amp_kw,
        bg_amp_bin=cfg["bg"]["amplitude binary threshold"],
        bg_amp_mask_radial_clearance=cfg["bg"]["amplitude mask sphere"],
        bg_pha_kw=bg_pha_kw,
        bg_pha_bin=cfg["bg"]["phase binary threshold"],
        bg_pha_mask_radial_clearance=cfg["bg"]["phase mask sphere"],
        bg_sphere_edge_kw=edge_kw,
        search_enabled=cfg["roi"]["enabled"],
        ret_roimgr=True,
        ret_changed=True,
    )
    if changed:
        print("Done.")
    else:
        print("Reusing.")
    if len(rmgr) == 0:
        print("No ROIs could be found!\n"
              + "Please try editing '{}':\n".format(cfg.path)
              + "- correct specimen size "
              + "({})\n".format(strpar(cfg, "specimen", "size um"))
              + "- increase allowed size variation "
              + "({})\n".format(strpar(cfg, "roi", "size variation"))
              + "- increase maximum allowed eccentricity "
              + "({})\n".format(strpar(cfg, "roi", "eccentricity max"))
              + "- reduce minimum distance to image border in pixels "
              + "({})\n".format(strpar(cfg, "roi", "dist border px"))
              + "- reduce minimum distance inbetween ROIs "
              + "({})".format(strpar(cfg, "roi", "exclude overlap px"))
              )
        sys.exit(1)

    if changed and cfg["output"]["roi images"]:
        print("Plotting detected ROIs... ", end="", flush=True)
        tifout = path_out / FILE_SENSOR_WITH_ROI_IMAGE
        # plot h5series and rmgr with matplotlib
        with qpimage.QPSeries(h5file=h5series, h5mode="r") as qps, \
                tifffile.TiffWriter(fspath(tifout), imagej=True) as tf:
            for ii in range(len(qps)):
                # new indexing convention in drymass 0.6.0
                image_index = ii + 1
                rois = rmgr.get_from_image_index(image_index)
                imio = io.BytesIO()
                plot.plot_qpi_phase(qps[ii],
                                    rois=rois,
                                    path=imio,
                                    labels_excluded=cfg["roi"]["ignore data"])
                imio.seek(0)
                imdat = (mpimg.imread(imio) * 255).astype("uint8")
                tf.save(imdat, compress=9)
        print("Done")

    if ret_data:
        return h5roi


def parse_bg_value(bg, reldir):
    """Determine the background to use from the configuration key"""
    if isinstance(bg, str):
        # can be a file name relative to the input directory
        # or an absolute path.
        reldir = pathlib.Path(reldir)
        path = reldir / bg
        if not path.exists():
            path = pathlib.Path(bg)
        bg = path
    return bg


def strpar(cfg, section, key):
    """String representation of a section/key combination"""
    return "[{}] {} = {}".format(section, key, cfg[section][key])



import numpy as np
import qpsphere

from drymass.anasphere import absolute_dry_mass_sphere, relative_dry_mass


def defaultsim(medium_index=1.335,
               m_g=150e-12):
    grid_size = (100, 100)
    center = (50, 50)
    wavelength = 550 * 1e-9
    pixel_size = .3 * 1e-6
    r_m = 10 * 1e-6
    alpha = .18
    alpha_m3g = alpha * 1e-6
    n = 1.335 + 3 * alpha_m3g * m_g / (4 * np.pi * (r_m**3))
    # generate example dataset
    qpi = qpsphere.simulate(radius=r_m,
                            sphere_index=n,
                            medium_index=medium_index,
                            wavelength=wavelength,
                            pixel_size=pixel_size,
                            model="projection",
                            grid_size=grid_size,
                            center=center)
    kwargs = {"qpi": qpi,
              "radius": r_m,
              "center": center,
              "alpha": alpha,
              "rad_fact": 1.2
              }
    return kwargs


def test_equivalence():
    m_g = 150e-12
    kwargs = defaultsim(m_g=m_g)
    mabs = absolute_dry_mass_sphere(**kwargs)
    mrel = relative_dry_mass(**kwargs)
    assert mabs == mrel
    assert np.allclose(mabs, m_g, atol=0, rtol=.0005)


def test_higher_medium():
    for m_g in np.linspace(30e-12, 150e-12, 4):
        for medium_index in np.linspace(1.334, 1.34, 8):
            kwargs = defaultsim(m_g=m_g, medium_index=medium_index)
            mabs = absolute_dry_mass_sphere(**kwargs)
            # accuracy depends on image resolution
            assert np.allclose(mabs, m_g, atol=0, rtol=.002)


def test_negative():
    m_g = 5e-12
    medium_index = 1.336
    kwargs = defaultsim(m_g=m_g, medium_index=medium_index)
    mrel = relative_dry_mass(**kwargs)
    mabs = absolute_dry_mass_sphere(**kwargs)
    assert mrel < 0
    assert np.allclose(mabs, m_g, atol=0, rtol=.002)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import shutil

from drymass import cli
import matplotlib.pylab as plt
import qpimage


path = "QLSR_PAA_beads_modified.h5"

# setup output directory
path_in, path_out = cli.dialog.main(path)
shutil.rmtree(path_out)


# Select 2nd order polynomial background correction
# -------------------------------------------------

# initial run with standard parameters
h5roi = cli.cli_extract_roi(path=path_in, ret_data=True)
h5sim = cli.cli_analyze_sphere(path=path_in, ret_data=True)

# get qpimage data of phase images with quadratic background
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5sim, h5mode="r") as qpsim:
    qe1 = [qproi[3].pha - qpsim[3].pha,
           qproi[4].pha - qpsim[4].pha]

# get configuration handle
cfg = cli.config.ConfigFile(path_out)

kwerr = {"px_um": cfg["meta"]["pixel size um"],
         "imtype": "phase error",
         "vmin": -0.5,
         "vmax": +0.5,
         "cbformat": "%.1f",
         }

kwplt = {"px_um": cfg["meta"]["pixel size um"]}

# repeat with poly2o correction
cfg.set_value("bg", "phase profile", "poly2o")
cli.cli_analyze_sphere(path=path_in, ret_data=True)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5sim, h5mode="r") as qpsim:
    qe2 = [qproi[3].pha - qpsim[3].pha,
           qproi[4].pha - qpsim[4].pha]

# repeat with no correction
cfg.set_value("bg", "enabled", False)
cli.cli_extract_roi(path=path_in)
cfg.set_value("bg", "enabled", True)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi:
    qd = [qproi[3].pha, qproi[4].pha]

# plot differences
fig, axes = plt.subplots(2, 3, figsize=(8, 4.4))
cli.plot.plot_image(qd[0], ax=axes[0, 0], **kwplt)
cli.plot.plot_image(qe1[0], ax=axes[0, 1], **kwerr)
cli.plot.plot_image(qe2[0], ax=axes[0, 2], **kwerr)

cli.plot.plot_image(qd[1], ax=axes[1, 0], **kwplt)
cli.plot.plot_image(qe1[1], ax=axes[1, 1], **kwerr)
cli.plot.plot_image(qe2[1], ax=axes[1, 2], **kwerr)

for ii in range(axes.shape[0]):
    axes[ii, 0].set_title("input phase")
    axes[ii, 1].set_title("tilt residuals")
    axes[ii, 2].set_title("poly2o residuals")

    axes[ii, 1].set_ylabel("")
    axes[ii, 2].set_ylabel("")

axes[0, 0].set_xlabel("")
axes[0, 1].set_xlabel("")
axes[0, 2].set_xlabel("")

plt.tight_layout(pad=0, h_pad=.6)
plt.savefig("_t03_quadratic_correction.jpg")
plt.close()

# Include beads that are close to each other
# ------------------------------------------
cfg.set_value("roi", "exclude overlap px", 0)
cfg.set_value("bg", "enabled", False)
cli.cli_extract_roi(path=path_in)
cfg.set_value("bg", "enabled", True)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi:
    td = [qproi[5].pha, qproi[7].pha]

cli.cli_analyze_sphere(path=path_in)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5sim, h5mode="r") as qpsim:
    te1 = [qproi[5].pha - qpsim[5].pha,
           qproi[7].pha - qpsim[7].pha]

cfg.set_value("bg", "phase binary threshold", "threshold_triangle")
cli.cli_analyze_sphere(path=path_in)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5sim, h5mode="r") as qpsim:
    te2 = [qproi[5].pha - qpsim[5].pha,
           qproi[7].pha - qpsim[7].pha]

# plot differences
fig, axes = plt.subplots(2, 3, figsize=(8, 4.4))
cli.plot.plot_image(td[0], ax=axes[0, 0], **kwplt)
cli.plot.plot_image(te1[0], ax=axes[0, 1], **kwerr)
cli.plot.plot_image(te2[0], ax=axes[0, 2], **kwerr)

cli.plot.plot_image(td[1], ax=axes[1, 0], **kwplt)
cli.plot.plot_image(te1[1], ax=axes[1, 1], **kwerr)
cli.plot.plot_image(te2[1], ax=axes[1, 2], **kwerr)

for ii in range(axes.shape[0]):
    axes[ii, 0].set_title("input phase")
    axes[ii, 1].set_title("residuals, no mask")
    axes[ii, 2].set_title("residuals, with mask")
    axes[ii, 1].set_ylabel("")
    axes[ii, 2].set_ylabel("")

axes[0, 0].set_xlabel("")
axes[0, 1].set_xlabel("")
axes[0, 2].set_xlabel("")

plt.tight_layout(pad=0, h_pad=.6)
plt.savefig("_t03_bead_overlap.jpg")
plt.close()


# Include beads at the border of the sensor image
# -----------------------------------------------
cfg.set_value("roi", "dist border px", 0)
cfg.set_value("bg", "enabled", False)
cli.cli_extract_roi(path=path_in)
cfg.set_value("bg", "enabled", True)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi:
    bd = qproi[9].pha

cli.cli_analyze_sphere(path=path_in)
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5sim, h5mode="r") as qpsim:
    be = qproi[9].pha - qpsim[9].pha


# plot differences
fig, axes = plt.subplots(1, 2, figsize=(5, 2.4))
cli.plot.plot_image(bd, ax=axes[0], **kwplt)
cli.plot.plot_image(be, ax=axes[1], **kwerr)

axes[0].set_title("input phase")
axes[1].set_title("phase residuals")
axes[1].set_ylabel("")

plt.tight_layout(pad=0)
plt.savefig("_t03_bead_border.jpg")
plt.close()

# Exact determination of radius and refractive index
# --------------------------------------------------
cfg.set_value("sphere", "method", "image")
cfg.set_value("sphere", "model", "rytov-sc")
h5ryt = cli.cli_analyze_sphere(path=path_in, ret_data=True)

errs = []
with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qproi, \
        qpimage.QPSeries(h5file=h5ryt, h5mode="r") as qpsim:
    for ii in range(len(qproi)):
        errs.append(qproi[ii].pha - qpsim[ii].pha)

errs.pop(8)

# plot differences
fig, axes = plt.subplots(3, 3, figsize=(8, 8))
axes = axes.flatten()
titles = ["1. offset",
          "2. linear tilt",
          "3. linear tilt",
          "4. quadratic distortion",
          "5. quadratic distortion",
          "6.1 two beads",
          "6.2 two beads",
          "7. all of the above",
          "8. bead at border",
          ]

for ii in range(9):
    cli.plot.plot_image(errs[ii], ax=axes[ii], **kwerr)
    axes[ii].set_title(titles[ii])

for ii in range(6):
    axes[ii].set_xlabel("")

for ii in [1, 2, 4, 5, 7, 8]:
    axes[ii].set_ylabel("")

plt.tight_layout(pad=0, h_pad=.6, w_pad=.6)
plt.savefig("_t03_summary_rytov-sc.jpg")
plt.close()



import pathlib
import warnings


class ROIManagerWarning(UserWarning):
    """Used for unexpected keyword arguments."""
    pass


class ROI(object):
    def __init__(self, identifier, image_index, roi_index, roi_slice):
        """Handle one region of interest (ROI)"""
        if not isinstance(identifier, str):
            raise ValueError("`identifier` must be a string!")
        # verify roi_slice
        if not (isinstance(roi_slice, (tuple, list))
                and len(roi_slice) == 2
                and isinstance(roi_slice[0], slice)
                and isinstance(roi_slice[1], slice)):
            raise ValueError("`roi_slice` must be a tuple of two slices!")

        self.identifier = identifier
        self.image_index = image_index
        self.roi_index = roi_index
        self.roi_slice = roi_slice

    def __eq__(self, other):
        return self.to_str() == other.to_str()

    def __lt__(self, other):
        # used for sorting
        if self.image_index == other.image_index:
            return self.roi_index < other.roi_index
        else:
            return self.image_index < other.image_index

    def __repr__(self):
        return self.to_str()

    @staticmethod
    def from_str(strin):
        """Load ROI from a string"""
        ll = strin.strip().split("\t")
        # extract slice data
        sldata = "".join([c for c in ll[3] if c in ",0123456789"])
        sldata = sldata.replace(",,", ",").split(",")
        sldata = [int(s) for s in sldata if s]
        slice1 = slice(min(sldata[0], sldata[1]),
                       max(sldata[0], sldata[1]))
        slice2 = slice(min(sldata[2], sldata[3]),
                       max(sldata[2], sldata[3]))
        roi = ROI(identifier=ll[0],
                  image_index=int(ll[1]),
                  roi_index=int(ll[2]),
                  roi_slice=(slice1, slice2)
                  )
        return roi

    def to_str(self):
        """Export ROI to a string"""
        strout = "\t".join([self.identifier,
                            str(self.image_index),
                            str(self.roi_index),
                            str(self.roi_slice)
                            ])
        return strout


class ROIManager(object):
    def __init__(self, identifier=None):
        """Manage regions of interest (ROI) of an image series

        Parameters
        ----------
        identifier: str or None
            The identifier of the image series.
        """
        if not (isinstance(identifier, str) or identifier is None):
            raise ValueError("`identifier` must be `None` or a string!")
        self.identifier = identifier
        self.rois = []

    def __len__(self):
        return len(self.rois)

    def add(self, roi_slice, image_index, roi_index, identifier):
        """Add a ROI to ROIManager

        Parameters
        ----------
        roi_slice: (slice, slice)
            Two slices indicating the x- and y-axis limits.
        image_index: int
            The image index of the image series.
        roi_index: int
            The ROI index in image `image_index`
        identifier: str
            The ROI identifier. If `self.identifier` is not contained
            within `identifier`, a `ROIManagerWarning` will be issued.
        """
        # verify identifier
        if not isinstance(identifier, str):
            raise ValueError("`identifier` must be a string!")
        if self.identifier and self.identifier not in identifier:
            msg = "Identifier of ROIManager `{}` ".format(self.identifier) \
                  + "does not match that of QPSeries `{}`.".format(identifier)
            warnings.warn(msg, ROIManagerWarning)
        self.rois.append(ROI(identifier=identifier,
                             image_index=image_index,
                             roi_index=roi_index,
                             roi_slice=roi_slice))

    def get_from_image_index(self, image_index):
        rois = [r for r in self.rois if r.image_index == image_index]
        return sorted(rois)

    def load(self, path):
        """Load ROIs from a text file"""
        path = pathlib.Path(path)
        with path.open(mode="r") as fd:
            lines = fd.readlines()
        for ll in lines:
            # ignore empty lines
            if ll.strip():
                self.rois.append(ROI.from_str(ll))

    def save(self, path):
        """Save ROIs to a text file (`path` will be overridden)"""
        path = pathlib.Path(path)
        with path.open(mode="w") as fd:
            for roi in self.rois:
                fd.write(roi.to_str() + "\r\n")



import pathlib
import shutil
import tempfile

import numpy as np
import qpimage

from drymass.cli import cli_convert, config, dialog


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1, write_config=True):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path_in = tempfile.mktemp(suffix=".h5", prefix="drymass_test_cli_convert")
    path_in = pathlib.Path(path_in)
    with qpimage.QPSeries(h5file=path_in, h5mode="w", identifier="tes") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    path_out = path_in.with_name(path_in.name + dialog.OUTPUT_SUFFIX)
    path_out.mkdir()
    if write_config:
        # add drymass configuration file
        cfg = config.ConfigFile(path_out)
        cfg.set_value(section="meta", key="pixel size um", value=pxsize*1e6)
        cfg.set_value(section="meta", key="wavelength nm",
                      value=wavelength*1e9)
        cfg.set_value(section="meta", key="medium index", value=medium_index)
    return qpi, path_in, path_out


def test_base():
    qpi, path_in, path_out = setup_test_data(num=2)
    h5data = cli_convert(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert np.allclose(qpi.pha, qps[0].pha)
        assert np.allclose(qpi.amp, qps[0].amp)

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_bg_corr_file():
    _bgqpi, bg_path_in, bg_path_out = setup_test_data(num=1)
    _qpi, path_in, path_out = setup_test_data(num=2)
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="bg", key="phase data", value=bg_path_in)
    cfg.set_value(section="bg", key="amplitude data", value=bg_path_in)
    h5data = cli_convert(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    try:
        path_in.unlink()
        bg_path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)
    shutil.rmtree(bg_path_out, ignore_errors=True)


def test_bg_corr_file_relative():
    _bgqpi, bg_path_in, bg_path_out = setup_test_data(num=1)
    _qpi, path_in, path_out = setup_test_data(num=2)
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="bg", key="phase data",
                  value=str(bg_path_in.relative_to(path_in.parent)))
    cfg.set_value(section="bg", key="amplitude data",
                  value=str(bg_path_in.relative_to(path_in.parent)))
    h5data = cli_convert(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    try:
        path_in.unlink()
        bg_path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)
    shutil.rmtree(bg_path_out, ignore_errors=True)


def test_bg_corr_index():
    _qpi, path_in, path_out = setup_test_data(num=2)
    cfg = config.ConfigFile(path_out)
    cfg.set_value(section="bg", key="phase data", value=1)
    cfg.set_value(section="bg", key="amplitude data", value=1)
    h5data = cli_convert(path=path_in, ret_data=True)

    with qpimage.QPSeries(h5file=h5data, h5mode="r") as qps:
        assert np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


def test_extract_meta_data():
    """qpformat can automatically transfer the metadata from
    QPSeries files to the dataset which DryMass makes use of.
    """
    _qpi, path_in, path_out = setup_test_data(pxsize=1.3e-6,
                                              medium_index=1.345,
                                              wavelength=555e-9,
                                              num=2,
                                              write_config=False)

    cli_convert(path=path_in)
    cfg = config.ConfigFile(path_out)

    assert cfg["meta"]["medium index"] == 1.345
    assert cfg["meta"]["pixel size um"] == 1.3
    assert cfg["meta"]["wavelength nm"] == 555

    try:
        path_in.unlink()
    except OSError:
        pass
    shutil.rmtree(path_out, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
import tempfile
import shutil

import numpy as np
import qpimage

import drymass


def setup_test_data(radius_px=30, size=200, pxsize=1e-6, medium_index=1.335,
                    wavelength=550e-9, num=1):
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    cx = 80
    cy = 120
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    pha = (r < radius_px) * 1.3
    amp = .5 + np.roll(pha, 10) / pha.max()
    qpi = qpimage.QPImage(data=(pha, amp),
                          which_data="phase,amplitude",
                          meta_data={"pixel size": pxsize,
                                     "medium index": medium_index,
                                     "wavelength": wavelength})
    path = tempfile.mktemp(suffix=".h5", prefix="drymass_test_convert")
    dout = tempfile.mkdtemp(prefix="drymass_test_sphere_")
    with qpimage.QPSeries(h5file=path, h5mode="w") as qps:
        for ii in range(num):
            qps.add_qpimage(qpi, identifier="test_{}".format(ii))
    return qpi, path, dout


def test_bg_correction_index():
    qpi, path, dout = setup_test_data(num=2)

    path_out = drymass.convert(path_in=path,
                               dir_out=dout,
                               bg_data_amp=1,
                               bg_data_pha=1)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qps:
        # background correction with same input image will result
        # in a flat QPImage.
        assert np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    # To be absolutely sure this works, append a blank
    # QPImage and do it again.
    with qpimage.QPSeries(h5file=path, h5mode="a") as qps:
        pha = .5 * np.ones(qps[0].shape)
        amp = .9 * np.ones(qps[0].shape)
        qps.add_qpimage(qpimage.QPImage(data=(pha, amp),
                                        which_data="phase,amplitude"))

    path_out = drymass.convert(path_in=path,
                               dir_out=dout,
                               bg_data_amp=3,
                               bg_data_pha=3)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qps:
        # background correction with same input image will result
        # in a flat QPImage.
        assert np.allclose(qps[0].pha, qpi.pha - .5)
        assert np.allclose(qps[0].amp, qpi.amp / .9)

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_correction_index_bad():
    _qpi, path, dout = setup_test_data(num=2)

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_amp=0)
    except ValueError:
        pass
    else:
        assert False

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_pha=-1)
    except ValueError:
        pass
    else:
        assert False

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_amp=3)
    except ValueError:
        pass
    else:
        assert False

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_pha=3)
    except ValueError:
        pass
    else:
        assert False

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_correction_invalid():
    _qpi, path, dout = setup_test_data(num=2)

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_pha=np.zeros(10))
    except ValueError:
        pass
    else:
        assert False

    try:
        drymass.convert(path_in=path, dir_out=dout, bg_data_pha=1.2)
    except ValueError:
        pass
    else:
        assert False

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_correction_none_bad():
    _qpi, path, dout = setup_test_data(num=2)

    path_out = drymass.convert(path_in=path, dir_out=dout, bg_data_amp=1)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qps:
        # background correction with same input image will result
        # in a flat QPImage.
        assert not np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    path_out = drymass.convert(path_in=path, dir_out=dout, bg_data_pha=1)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qps:
        # background correction with same input image will result
        # in a flat QPImage.
        assert np.all(qps[0].pha == 0)
        assert not np.all(qps[0].amp == 1)

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_bg_correction_path():
    _qpi, path, dout = setup_test_data(num=2)
    _bgqpi, bgpath, bgdout = setup_test_data(num=1)

    path_out = drymass.convert(path_in=path, dir_out=dout,
                               bg_data_pha=bgpath,
                               bg_data_amp=bgpath)
    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qps:
        # background correction with same input image will result
        # in a flat QPImage.
        assert np.all(qps[0].pha == 0)
        assert np.all(qps[0].amp == 1)

    try:
        os.remove(path)
        os.remove(bgpath)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)
    shutil.rmtree(bgdout, ignore_errors=True)


def test_change_wavelength():
    _qpi, path, dout = setup_test_data()
    path_out = drymass.convert(path_in=path,
                               dir_out=dout,
                               meta_data={"wavelength": 500e-9})

    with qpimage.QPSeries(h5file=path_out, h5mode="r") as qpso:
        id1 = qpso.identifier

    path_out2 = drymass.convert(path_in=path,
                                dir_out=dout,
                                meta_data={"wavelength": 333e-9})

    with qpimage.QPSeries(h5file=path_out2, h5mode="r") as qpso:
        id2 = qpso.identifier

    assert id1 != id2, "Files should have different identifiers"

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


def test_reuse():
    _qpi, path, dout = setup_test_data(num=2)

    _po, changed1 = drymass.convert(path_in=path,
                                    dir_out=dout,
                                    bg_data_amp=1,
                                    bg_data_pha=1,
                                    ret_changed=True)
    _po, changed2 = drymass.convert(path_in=path,
                                    dir_out=dout,
                                    bg_data_amp=1,
                                    bg_data_pha=1,
                                    ret_changed=True)
    assert changed1
    assert not changed2

    try:
        os.remove(path)
    except OSError:
        pass
    shutil.rmtree(dout, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import warnings

import numpy as np
import qpimage
import qpsphere

from . import util

#: Output sphere analysis qpimage.QPSeries data
FILE_SPHERE_DATA = "sphere_{}_{}_data.h5"
#: Output sphere analysis statistics
FILE_SPHERE_STAT = "sphere_{}_{}_statistics.txt"


class EdgeDetectionFailedWarning(UserWarning):
    pass


def analyze_sphere(h5roi, dir_out, r0=10e-6, method="edge",
                   model="projection", edgekw={}, imagekw={},
                   alpha=.18, rad_fact=1.2, ret_changed=False):
    """Perform sphere analysis

    Parameters
    ----------
    h5series: str
        Path of qpimage.QPSeries hdf5 file
    dir_out: str
        Path to output directory
    r0: float
        Initial radius
    method: str
        Either "edge" or "image"; see :ref:`config_sphere` for
        more information.
    model: str
        Propagation model to use; see :ref:`config_sphere` for
        more information.
    edgekw: dict
        Keyword arguments to :func:`qpsphere.edgefit.contour_canny`
    imagekw: dict
        Keyword arguments to :func:`qpsphere.imagefit.alg.match_phase`
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Radial inclusion factor for dry mass computation
    ret_changed: bool
        Return boolean indicating whether the sphere data on disk was
        created/updated (True) or whether previously created ROI
        data was used (False).
    """
    dir_out = pathlib.Path(dir_out).resolve()

    h5out = dir_out / FILE_SPHERE_DATA.format(method, model)
    statout = dir_out / FILE_SPHERE_STAT.format(method, model)

    with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps:
        cfgid = util.hash_object([qps,
                                  r0,
                                  method,
                                  model,
                                  edgekw,
                                  imagekw if method == "image" else None,
                                  alpha,
                                  rad_fact])

    create = mode_for_sphere_analysis(h5in=h5roi, h5out=h5out, cfgid=cfgid)

    # initialize file
    initmode = "w" if create else "r"
    with qpimage.QPSeries(h5file=h5out, h5mode=initmode) as qps_out:
        # get all simulation identifiers from previous analysis
        ids_out = [qpi["identifier"] for qpi in qps_out]

    with qpimage.QPSeries(h5file=h5roi, h5mode="r") as qps_in, \
            statout.open(mode="w") as fd:

        header = ["identifier",
                  "index",
                  "radius_um",
                  "rel_dry_mass_pg",
                  "abs_dry_mass_pg",
                  "time",
                  "medium",
                  ]
        fd.write("#" + "\t".join(header) + "\r\n")

        for qpi in qps_in:
            simident = "{}:{}:sim:{}".format(qpi["identifier"], cfgid, model)
            if simident in ids_out:
                # read simulation results
                with qpimage.QPSeries(h5file=h5out, h5mode="r") as qps_out:
                    n = qps_out[simident]["sim index"]
                    r = qps_out[simident]["sim radius"]
                    c = qps_out[simident]["sim center"]
            else:
                try:
                    # fit sphere model
                    n, r, c, qpi_sim = qpsphere.analyze(qpi,
                                                        r0=r0,
                                                        method=method,
                                                        model=model,
                                                        edgekw=edgekw,
                                                        imagekw=imagekw,
                                                        ret_center=True,
                                                        ret_qpi=True)
                except qpsphere.edgefit.RadiusExceedsImageSizeError:
                    # Edge detection cannot proceed because the presumed
                    # object radius exceeds the image size. This might be
                    # the result of a "size variation" set too large.
                    msg = "Edge detection failed for ROI " \
                          + "{}! Try ".format(qpi["identifier"]) \
                          + "reducing the value of [roi]: 'size variation'."
                    warnings.warn(msg, EdgeDetectionFailedWarning)
                    # use dummy data
                    n = np.nan
                    r = np.nan
                    c = (np.nan, np.nan)
                    qpi_sim = qpsphere.simulate(radius=0,
                                                sphere_index=1,
                                                grid_size=qpi.shape)
                # write simulation results
                with qpimage.QPSeries(h5file=h5out, h5mode="a") as qps_out:
                    qps_out.add_qpimage(qpi=qpi_sim, identifier=simident)
            # finally, update text file
            if "time" in qpi:
                qptime = qpi["time"]
            else:
                qptime = 0
            data = {
                "identifier": qpi["identifier"],
                "index": n,
                "radius_um": r * 1e6,
                "abs_dry_mass_pg": absolute_dry_mass_sphere(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "rel_dry_mass_pg": relative_dry_mass(
                    qpi=qpi,
                    radius=r,
                    center=c,
                    alpha=alpha,
                    rad_fact=rad_fact
                ) * 1e12,
                "time": qptime,
                "medium": qpi["medium index"]
            }
            fd.write("\t".join([str(data[k]) for k in header]) + "\r\n")
            fd.flush()

    if ret_changed:
        return h5out, create
    else:
        return h5out


def absolute_dry_mass_sphere(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute absolute dry mass of a spherical phase object

    The absolute dry mass is computed with

    .. code::

        m_abs = m_rel + m_sup
        m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA
        m_sup = 4*PI / (3*alpha) * radius^3 (n_med - n_PBS)

    with the vacuum wavelength ``lambda``, the total phase
    retardation in the area of interest ``phi_tot``, the pixel
    area ``deltaA``, the refractive index of the medium `n_med`
    (stored in ``qpi.meta``), and the refractive index of phosphate
    buffered saline (PBS) ``n_PBS=1.335``.

    This is the *absolute* dry mass, because it takes into account
    the offset caused by the suppressed density in the phase data.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the sphere [px]
    radius: float
        Radius of the sphere [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The absolute dry mass of the sphere [g]
    """
    dm_rel = relative_dry_mass(qpi=qpi,
                               radius=radius,
                               center=center,
                               alpha=alpha,
                               rad_fact=rad_fact)
    medium_index = qpi["medium index"]
    dm_sup = 4 / 3 * np.pi / (alpha * 1e-6) * \
        radius**3 * (medium_index - 1.335)
    return dm_rel + dm_sup


def relative_dry_mass(qpi, radius, center, alpha=.18, rad_fact=1.2):
    """Compute relative dry mass of a circular area in QPI

    The dry mass is computed with

    m_rel = lambda / (2*PI*alpha) * phi_tot * deltaA

    with the vacuum wavelength `lambda`, the total phase
    retardation in the area of interest `phi_tot`, and the pixel
    area `deltaA`.

    This is the *relative* dry mass, because it is computed relative
    to the surrounding medium (phi_tot) and not relative to water.

    Parameters
    ----------
    qpi: qpimage.QPImage
        QPI data
    center: tuble (x,y)
        Center of the area of interest [px]
    radius: float
        Radius of the area of interest [m]
    wavelength: float
        The wavelength of the light [m]
    alpha: float
        Refraction increment [mL/g]
    rad_fact: float
        Inclusion factor that scales `radius` to increase
        the area used for phase summation; if the backgound
        phase exhibits a noise phase signal, positive and
        negative contributions cancel out and `rad_fact`
        does not have an effect above a certain critical value.

    Returns
    -------
    dry_mass: float
        The relative dry mass of the object [g]
    """
    image = qpi.pha
    rincl = radius / qpi["pixel size"] * rad_fact
    wavelength = qpi["wavelength"]
    sx, sy = qpi.shape
    x = np.arange(sx).reshape(-1, 1)
    y = np.arange(sy).reshape(1, -1)
    discsq = ((x - center[0])**2 + (y - center[1])**2)
    area = discsq < rincl**2
    phi_tot = np.sum(image[area])
    # compute dry mass
    pxarea = qpi["pixel size"]**2
    # convert alpha mL/g to m³/g
    fact = 1e-6
    # [kg]
    dm = wavelength / (2 * np.pi * alpha * fact) * phi_tot * pxarea
    return dm


def mode_for_sphere_analysis(h5in, h5out, cfgid):
    """Determine the mode for the QPSeries file for subsequent analysis

    Sometimes an analysis is interrupted and the output files are
    still intact. This method determines whether it is possible to
    continue the analysis where left off or not.

    Parameters
    ----------
    h5in: pathlib.Path
        The input QPSeries file
    h5out: pathlib.Path
        The output QPSeries file
    cfgid: str
        The configuration hash of the sphere analysis which is
        part of the output QPSeries analysis

    Returns
    -------
    create: bool
        Whether the output QPSeries file is ok. This is dependent
        on the following scenarios:

        - True: There is no output QPSeries file, it is corrupt,
              or at least one of the qpimage identifiers is not present
              in the input QPSeries.
        - False: Some of the input QPSeries identifiers are
              present in the output QPSeries.
    """
    with qpimage.QPSeries(h5file=h5in, h5mode="r") as qps_in:
        # read all identifiers
        idsin = [qpi["identifier"] for qpi in qps_in]

    try:
        with qpimage.QPSeries(h5file=h5out, h5mode="r") as qps_out:
            # read all identifiers
            idsout = [qpi["identifier"] for qpi in qps_out]
    except (IOError, OSError):
        # corrupt file
        idsout = []

    valid = []  # matching identifiers
    bad = []  # those exist in h5out but not in h5in

    for oid in idsout:
        iid, iicfg = oid.rsplit(":", 3)[:2]
        if iid in idsin and iicfg == cfgid:
            valid.append(iid)
        else:
            bad.append(iid)

    return bad or not valid



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'user-friendly quantitative phase imaging analysis'
name = 'drymass'
year = "2017"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version

if version.count("dev") or sys.argv.count("test"):
    # specific versions are not desired for
    # - development version
    # - running pytest
    release_deps = ["qpformat",
                    "qpimage",
                    "qpsphere"]
else:
    release_deps = ["qpformat==0.4.2",
                    "qpimage==0.4.2",
                    "qpsphere==0.4.0",
                    ]

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/RI-imaging/DryMass',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["matplotlib",
                      "numpy>=1.12.0",
                      "scikit-image>=0.13.1",
                      ] + release_deps,
    setup_requires=['pytest-runner'],
    tests_require=["pytest"],
    python_requires='>=3.6, <4',
    entry_points={
       "console_scripts": [
           "dm_analyze_sphere = drymass.cli:cli_analyze_sphere",
           "dm_convert = drymass.cli:cli_convert",
           "dm_extract_roi = drymass.cli:cli_extract_roi",
            ],
       },
    keywords=["digital holographic microscopy",
              "optics",
              "quantitative phase imaging",
              "refractive index",
              "scattering",
              ],
    classifiers= [
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )



import pathlib
import tempfile
import shutil
import warnings

from drymass import roi


def test_basic():
    rmg = roi.ROIManager(identifier="test")
    assert len(rmg) == 0

    roi_slice = slice(4, 10), slice(5, 10)
    image_index = 2
    roi_index = 5
    identifier = "test_2_5"
    rmg.add(roi_slice, image_index, roi_index, identifier)
    assert len(rmg) == 1

    regions = rmg.get_from_image_index(image_index)
    reg = regions[0]
    assert reg.identifier == identifier
    assert reg.roi_slice == roi_slice


def test_save_load():
    rmg = roi.ROIManager(identifier="test")
    assert len(rmg) == 0

    roi_index = 5
    identifier = "test_2_5"
    for ii in [2, 5, 7]:
        image_index = ii
        identifier = "test_{}".format(image_index)
        roislice = slice(4+ii, 10+ii), slice(5, 10)
        rmg.add(roislice, image_index, roi_index, identifier)

    tdir = tempfile.mkdtemp(prefix="test_drymass_roi_manager_")
    path = pathlib.Path(tdir) / "test_roi.txt"
    rmg.save(path)

    rmg2 = roi.ROIManager(identifier="test_")
    rmg2.load(path)

    assert len(rmg) == len(rmg2)

    for ii in [2, 5, 7]:
        assert rmg2.get_from_image_index(ii) == rmg.get_from_image_index(ii)

    shutil.rmtree(tdir, ignore_errors=True)


def test_valueerror():
    try:
        roi.ROIManager(identifier=2)
    except ValueError:
        pass
    else:
        assert False, "identifier must be string or None"

    rmg = roi.ROIManager()
    roislice = slice(4, 10), slice(5, 10)
    image_index = 2
    roi_index = 5
    identifier = "test_2_5"

    try:  # bad roislice
        rmg.add(roislice[0], image_index, roi_index, identifier)
    except ValueError:
        pass
    else:
        assert False, "bad roi slice"

    try:  # bad identifier
        rmg.add(roislice, image_index, roi_index, None)
    except ValueError:
        pass
    else:
        assert False, "bad identifier"


def test_identifier_warnging():
    rmg = roi.ROIManager(identifier="not_in_roi_slice")
    roislice = slice(4, 10), slice(5, 10)
    image_index = 2
    roi_index = 5
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        rmg.add(roislice, image_index, roi_index, "testA")
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, roi.ROIManagerWarning)
        assert "does not match that of QPSeries" in str(w[-1].message)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numbers
from os import fspath
import pathlib

import numpy as np


def fbool(value):
    """Boolean value from string or number"""
    if isinstance(value, str):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        elif value:
            value = bool(float(value))
        else:
            raise ValueError("Got empty string!")
    else:
        value = bool(float(value))
    return value


def fintlist(alist):
    """List of integers from string or list of strings/integers"""
    outlist = []
    if isinstance(alist, str):
        # we have a string (comma-separated integers)
        alist = alist.strip().strip("[] ").split(",")
    for it in alist:
        if it:
            outlist.append(int(float(it)))
    return outlist


def float01(flt):
    """Float value between 0 and 1"""
    flt = float(flt)
    if flt < 0 or flt > 1:
        raise ValueError("Input must be between 0 and 1!")
    return flt


def float_or_str(flt_or_str):
    """Float value from string or number"""
    if isinstance(flt_or_str, str):
        flt_or_str = flt_or_str.strip()
        if flt_or_str == "nan":
            return np.nan
        else:
            try:
                value = float(flt_or_str)
            except ValueError:
                return flt_or_str
            else:
                return value
    else:
        return float(flt_or_str)


def floattuple_or_one(fti):
    """Tuple of two floats or ±1 from a string or a number"""
    msg = "Expected +1, -1 or (float, float), got '{}'!".format(fti)
    if isinstance(fti, (list, tuple)):
        if len(fti) != 2:
            raise ValueError(msg)
        fti = [float(fti[0]), float(fti[1])]
    elif isinstance(fti, str):
        if fti.count(","):  # tuple
            for s in " ()[]":
                fti = fti.replace(s, "")
            fti = floattuple_or_one(fti.split(","))
        else:  # one
            try:
                fti = int(float(fti))
            except ValueError:
                raise ValueError(msg)
    elif isinstance(fti, numbers.Number):
        fti = int(fti)
        if fti in [+1, -1]:
            pass
        else:
            raise ValueError(msg)
    else:
        fti = floattuple_or_one(list(fti))
    return fti


def int_or_path(intpath):
    """Integer or string from a string or a number"""
    if isinstance(intpath, str):
        intpath = intpath.strip()
        if intpath.replace(".", "").isdigit():
            value = int(float(intpath))
        else:
            value = fspath(intpath)
    elif isinstance(intpath, pathlib.Path):
        value = fspath(intpath)
    else:
        value = int(float(intpath))
    return value


def lcstr(astr):
    """Convert a string to lower-case"""
    return astr.lower()


def strlist(alist):
    """List of strings, comma- or space-separated"""
    if isinstance(alist, str):
        for s in "()[]'"+'"':
            alist = alist.replace(s, "")
        alist = alist.replace(",", " ")
        alist = alist.split(" ")
    alist = [a for a in alist if a.strip()]
    alist = sorted(alist)
    return alist


def tupletupleint(items):
    """A tuple containing x- and y- slice tuples from a string or tuple"""
    if isinstance(items, str):
        for s in " ()[]":
            items = items.replace(s, "")
        # we have a string representation of the slices
        x1, x2, y1, y2 = items.split(",")
        out = ((int(float(x1)), int(float(x2))),
               (int(float(y1)), int(float(y2))))
    else:
        (x1, x2), (y1, y2) = items
        out = ((int(float(x1)), int(float(x2))),
               (int(float(y1)), int(float(y2))))
    return out


func_types = {fbool: bool,
              fintlist: list,
              lcstr: str,
              tupletupleint: tuple,
              }
